class MessagesController < ApplicationController
  helper_method :truncated_message_description
  
  MESSAGES_PER_PAGE = 10
  
  skip_before_filter :authenticate_user!, only: [:widget, :show]

  before_filter :ensure_owner_or_recommender, only: :delete_item
  
  before_filter :authorize_user!, only: [
    :add_comment, :change_featured_item, :like_item, :unlike_item, :leave_item, :unleave_item, :add_item,
    :update_last_visit_time, :delete_comment, :activities, :send_invitation, :edit, :update, :show
  ]

  protect_from_forgery except: [:add_item, :create, :like_item, :unlike_item, :save_item]
  
  def new
    @message = Message.new(public: true)
    
    respond_to do |format|
      format.js { render :layout => false }
    end
  end
  
  def edit
    # @message is prepared from "authorize_user!" filter
    
    respond_to do |format|
      format.js { render :layout => false }
    end
  end
  
  def update
    # @message is prepared from "authorize_user!" filter
    
    @message.attributes = params[:message]
    if @message.private?
      require_dependency 'emails_tokenizer'
      recipients = EmailsTokenizer.tokenize( (params[:recipients] || []).join(',') )
    end    
    
    Message.transaction do
      @message.save!
      if @message.private?
        @activities = @message.send_invitation_email(current_user, recipients)
      else
        @message.users.clear
      end
      
      # set the last visit to now
      current_user.visit_message!(@message)
    end
    
    respond_to do |format|
      format.js do
        render :json => {
          :activities_html => (@activities || []).select {|activity| activity.visible? }.collect {|activity|
            render_to_string({
              :partial => 'activity', :formats => 'html', :locals => {:activity => activity}
            })
          },
          :recipients_html => render_to_string(:partial => 'recipients', :locals => {:message => @message}),
          :recipients_count_html => render_to_string(:partial => 'recipients_count', :locals => {:message => @message}),
          :friends_overlay => render_to_string(:partial => 'friends_overlay', :locals => {:message => @message})
        }, :status => 200
      end
      
      format.json do
        render :json => {
          :activities_html => (@activities || []).select {|activity| activity.visible? }.collect {|activity|
            render_to_string({
              :partial => 'activity', :formats => 'html', :locals => {:activity => activity}
            })
          }
        }, :status => 200
      end
    end
  end
  
  def create
    @message = current_user.owned_messages.build( params[:message] )
    @message.image_uploaded = true if @message.image_file_name_changed?
    if @message.private?
      require_dependency 'emails_tokenizer'
      recipients = EmailsTokenizer.tokenize( (params[:recipients] || []).join(',') )
    end
    
    begin
      @message.save!
      if @message.private?
        @message.send_invitation_email(@message.owner, recipients)
      else
        @message.users.clear
      end
      
      respond_to do |format|
        format.js do
          render :json => {
            location: message_path(@message)
          }, :status => 200
        end
        format.extchr { render :text => @message.id }
      end
    rescue ActiveRecord::RecordInvalid => e
      respond_to do |format|
        format.js do
          render 'new', :layout => false, :status => 500
        end
        format.extchr { render :json => {:errors => @message.errors.full_messages} }
      end
    rescue Exception => e
      Airbrake.notify_or_ignore(e)
      
      respond_to do |format|
        format.js do
          render 'new', :layout => false, :status => 500
        end
        format.extchr { render :json => {:errors => @message.errors.full_messages} }
      end
    end
  end
  
  def open
    respond_to do |format|
      format.html { redirect_to my_profile_path }
      format.extchr do
        @active_messages = current_user.total_messages(true).non_featured_ones.active.order(
          'last_item_created_at DESC'
        )
        
        render json: construct_collections_list(@active_messages)
      end
    end
  end

  # construnct a list of ids and subjects out of messages array
  def construct_collections_list(messages)
    list = Array.new
    messages.each do |message|
      list << {id: message.id, subject: message.subject}
    end
    return list
  end

  def closed
    @inactive_messages = current_user.total_messages.inactive.order('last_activity_at DESC')
    
    @purchased_messages = Hash.new {|h,k| h[k] = [] }
    Item.purchased.where(:message_id => @inactive_messages.collect(&:id)).each do |purchased_item|
      @purchased_messages[ purchased_item.message_id ] << purchased_item
    end
    
    @messages_owners = {}
    User.where(:id => @inactive_messages.collect(&:owner_id).uniq).each do |user|
      @messages_owners[user.id] = user
    end
    
    respond_to do |format|
      format.html { render 'dashboard' }
    end
  end  

  # GET /messages/1
  # GET /messages/1.json
  def show
    @activities = @message.activities.visible.includes(:user).order('created_at desc').limit(MESSAGES_PER_PAGE)
    
    if current_user
      @visit = current_user.visits.where(:message_id => @message.id).first
    else
      @visit = nil
    end
    
    if @visit
      @unread_activities_ids = @activities.select {|activity| !activity.read_by?(current_user, @visit) }.collect(&:id)
    else
      @unread_activities_ids = @activities.collect(&:id)
    end

    @message_items = @message.items.order("created_at DESC")
    @featured_item = @message_items.select {|item| item.id == params[:featured_item].to_i }.first || @message_items[0]

    @original_content_activity = Activity.new
    @original_content_activity.user = @message.owner
    @original_content_activity.activity = @message.original_content
    @original_content_activity.created_at = @message.created_at
    
    # set the last visit to now
    current_user.visit_message(@message, @visit) if current_user

    impression = @message.impressions.where(user_id: current_user.try(:id)).first_or_initialize
    impression.site_impressions += 1
    impression.last_viewed_at = Time.now
    impression.last_view_source = 'site'
    impression.save!

    respond_to do |format|
      format.html
      format.json { render json: @item }
    end
  end

  def activities
    # @message is prepared from "authorize_user!" filter
    
    @activities = @message.activities.visible.includes(:user).order('created_at desc').
                           where("id < #{ params[:last_comment_id] }").limit(MESSAGES_PER_PAGE)
    
    respond_to do |format|
      format.js do
        render :json => {
          :activities_html => render_to_string({
            :partial => 'activities', :formats => 'html', :locals => {:activities => @activities}
          })
        }, :status => 200
      end
    end
  end

  def change_title
    @message = current_user.owned_messages.find(params[:id])
    @message.subject = params[:subject]
    
    if @message.save
      respond_to do |format|
        format.js do
          render :json => {
            :subject => @message.subject, :truncated_subject => view_context.truncate(@message.subject, :length => 70)
          }, :status => 200
        end
      end
    else
      respond_to do |format|
        format.js { render :nothing => true, :status => 500 }
      end
    end
  end
  def change_description
    @message = current_user.owned_messages.find(params[:id])
    @message.original_content = params[:description]
    
    if @message.save
      respond_to do |format|
        format.js do
          render :json => {
            :description => @message.original_content,
            :truncated_description => truncated_message_description(@message.original_content)
          }, :status => 200
        end
      end
    else
      respond_to do |format|
        format.js { render :nothing => true, :status => 500 }
      end
    end
  end

  # users involved in a message can send invitation to other users to get on message
  def send_invitation
    # @message is prepared from "authorize_user!" filter
    
    if @message.owner_id == current_user.id
      @message.attributes = params[:message]
      @message.image_uploaded = true if @message.image_file_name_changed?
    end
    if @message.private?
      require_dependency 'emails_tokenizer'
      recipients = EmailsTokenizer.tokenize( (params[:recipients] || []).join(',') )
    end
    
    Message.transaction do
      @message.save!
      if @message.private?
        @activities = @message.send_invitation_email(current_user, recipients)
      else
        @message.users.clear
      end
      
      # set the last visit to now
      current_user.visit_message!(@message)
    end
    
    respond_to do |format|
      format.js do
        render :json => {
          :activities_html => (@activities || []).select {|activity| activity.visible? }.collect {|activity|
            render_to_string({
              :partial => 'activity', :formats => 'html', :locals => {:activity => activity}
            })
          },
          :recipients_html => render_to_string(:partial => 'recipients', :locals => {:message => @message}),
          :recipients_count_html => render_to_string(:partial => 'recipients_count', :locals => {:message => @message}),
          :friends_overlay => render_to_string(:partial => 'friends_overlay', :locals => {:message => @message})
        }, :status => 200
      end
      
      format.json do
        render :json => {
          :activities_html => (@activities || []).select {|activity| activity.visible? }.collect {|activity|
            render_to_string({
              :partial => 'activity', :formats => 'html', :locals => {:activity => activity}
            })
          }
        }, :status => 200
      end
    end
  end
  
  def accept_invitation
    @message = current_user.involved_messages.find(params[:id])
    
    # User is considered friend with inviter from the time the invitation is sent with no need to accept
    #current_user.friends << @message.owner unless current_user.total_friends.include?(@message.owner)
    
    # current_user.activities.create(message_id: @message.id, kind: Activity::ACCEPT_INVITATION,
                                   # activity: I18n.t('messages.user_joined', name: current_user.name))
    
    redirect_to message_path(@message.id, :new_user => 1)
  end

  def remove_user
    @message = current_user.owned_messages.find(params[:id])
    
    @message.users.delete( User.find(params[:user_id]) )
    
    respond_to do |format|
      format.js { render :nothing => true, :status => 200 }
    end
  end

  # GET "/messages/1/change_featured_item?item=1"
  def change_featured_item
    # @message is prepared from "authorize_user!" filter
    
    @item = @message.items.unscoped.find(params[:item])
    
    respond_to do |format|
      format.js
      
      format.json do
        render :json => {
          :featured_item_html => render_to_string({
            :partial => 'featured_item',
            :formats => 'html',
            :locals => {featured_item: @item}
          }),
        }, :status => 200
      end
    end
  end

  def delete_item
    # @message & @item are prepared from "ensure_owner_or_recommender" filter
    
    @item.update_attribute(:deleted, true)
    
    @selected_item = @message.items.first
    
    respond_to do |format|
      format.js { render :nothing => true, :status => 200 }
      format.json do
        render :json => {
          :featured_item_html => (@message.items.empty? ? "" : render_to_string({
            :partial => 'featured_item',
            :formats => 'html',
            :locals => {featured_item: @selected_item}
          })),
          :item_id => @item.id
        }, :status => 200
      end
    end
  end

  def save_item
    # @message is prepared from "authorize_user!" filter
    
    collection = current_user.saved_items_collection
    item = Item.find_by_id( params[:item] )
    item.copy_to_collection!(collection.id, current_user) if item
    
    respond_to do |format|
      format.js { render nothing: true, status: 200 }
    end
  end 

  def like_item
    # @message is prepared from "authorize_user!" filter
    
    unless current_user.like?(params[:id], params[:item])
      @message, @item, @like = current_user.like(@message, params[:item])
    end
    
    respond_to do |format|
      format.js { render nothing: true, status: 200 }
      
      format.json do
        render :json => {
          :item_id => @item.id,
          :item_like_actions_html => render_to_string({
            :partial => 'item_like_actions',
            :formats => 'html',
            :locals => {item: @item}
          }),
          :activity_html => '' || render_to_string({
            :partial => 'activity',
            :formats => 'html',
            :locals => {activity: @like}
          })
        }, :status => 200
      end
    end
  end

  def unlike_item
    # @message is prepared from "authorize_user!" filter
    
    if current_user.like?(params[:id], params[:item])
      @message, @item, @activity_id = current_user.unlike(@message, params[:item])
    end
    
    respond_to do |format|
      format.js { render nothing: true, status: 200 }
      
      format.json do
        render :json => {
          :item_id => @item.id,
          :item_like_actions_html => render_to_string({
            :partial => 'item_like_actions',
            :formats => 'html',
            :locals => {item: @item}
          }),
          :activity_id => @activity_id
        }, :status => 200
      end
    end
  end
  
  def leave_item
    # @message is prepared from "authorize_user!" filter
    
    unless current_user.hate?(params[:id], params[:item])
      @message, @item, @leave = current_user.leave(@message, params[:item])
    end
    
    respond_to do |format|
      format.js
      
      format.json do
        render :json => {
          :leaves_count => @item.leaves_count,
          :likes_count => @item.likes_count,
          :item_id => @item.id,
          :item_like_actions_html => render_to_string({
            :partial => 'item_like_actions',
            :formats => 'html',
            :locals => {item: @item}
          }),
          :activity_html => '' || render_to_string({
            :partial => 'activity',
            :formats => 'html',
            :locals => {activity: @leave}
          })
        }, :status => 200
      end
    end
  end

  def unleave_item
    # @message is prepared from "authorize_user!" filter
    
    if current_user.hate?(params[:id], params[:item])
      @message, @item, @activity_id = current_user.unleave(@message, params[:item])
    end
    
    respond_to do |format|
      format.js
      
      format.json do
        render :json => {
          :leaves_count => @item.leaves_count,
          :likes_count => @item.likes_count,
          :item_id => @item.id,
          :item_like_actions_html => render_to_string({
            :partial => 'item_like_actions',
            :formats => 'html',
            :locals => {item: @item}
          }),
          :activity_id => @activity_id
        }, :status => 200
      end
    end
  end
  
  # POST /messages/:id/add_item.js
  def add_item
    # @message is prepared from "authorize_user!" filter
    
    @item = @message.items.build
    if params[:format] == 'extchr'
      @item.url = params[:item][:url]
      @item.title = params[:item][:title]
      @item.processed = false
    else
      @item.attributes = params[:item]
    end
    @item.user = current_user

    begin
      @item.save!
      
      if params[:format] == 'extchr'
        Background::ItemDataAdjuster.perform(@item.id, params[:item][:image_url])
      end
      
      @activity = current_user.activities.create({
        message_id: @message.id, item_id: @item.id,
        kind: Activity::CREATE_ITEM,
        activity: I18n.t('messages.item_created', name: current_user.first_name,
                                                  item: @item.title,
                                                  brand: @item.brand),
        performed_by: current_user,
        visible: false
      })
      
      respond_to do |format|
        format.js do
          render partial: 'messages/items', locals: {items: [@item]}, :status => 200
        end
        
        format.extchr { render text: 'Item was added' }
      end
    rescue ActiveRecord::RecordInvalid => e
      respond_to do |format|
        format.js do
          render :json => {:errors => @item.errors.full_messages}, :status => 406
        end
        format.extchr do
          render json: {errors: @item.errors.full_messages}, status: 406
        end
      end
    end
  end
  
  def add_comment
    # @message is prepared from "authorize_user!" filter
    
    @comment = current_user.activities.create!({
      message_id: @message.id,
      kind: Activity::COMMENT,
      activity: params[:activity][:activity],
      item_id: params[:activity][:item_id],
      performed_by: current_user
    })
    
    respond_to do |format|
      format.js
      
      format.json do
        response = {
          :comment_html => render_to_string({
            :partial => 'messages/activities', :formats => ['html'], :locals => {:activities => [@comment]}
          })
        }
        if @comment.item_id.present?
          response[:item_comments_count] = @comment.item.comments_count
        end
        
        render :json => response, :status => 200
      end
    end
  end

  # DELETE /messages/:id/delete_comment.js
  def delete_comment
    # @message is prepared from "authorize_user!" filter
    
    @comment = Activity.find( params[:comment_id] )
    unless (@comment.performed_by_id == current_user.id) || (@message.owner_id == current_user.id)
      raise ActiveRecord::RecordNotFound.new
    end 
    
    # delete the activity if it is of type comment
    if @comment.comment?
      @comment_id = @comment.id 
      @comment.destroy
    end

    respond_to do |format|
      format.js
      
      format.json do
        render :json => {}, :status => 200
      end
    end
  end
  
   # GET /messages/:id/select_items.js
  def select_items
    @message = current_user.owned_messages.find params[:id]
    @items = @message.items
    
    respond_to do |format|
      format.html
      format.js { render 'select_items', :formats => 'html', :layout => false }
    end
  end

  def close
    @message = current_user.owned_messages.private_ones.find( params[:id] )
    @message.update_attribute(:active, false);
    @message.items.update_all({purchased: true}, {id: params[:items]})
    
    respond_to do |format|
      format.html
      format.js { render :json => {}, :status => 200 }
    end
  end

  # PUT /messages/:id/update_last_visit_time.js
  def update_last_visit_time
    # @message is prepared from "authorize_user!" filter
    
    @unread_activities = current_user.unread_activities(@message).visible.order('created_at desc')
    @unread_activities_ids = @unread_activities.map(&:id)

    # set the last visit to now
    current_user.visit_message(@message)
    
    respond_to do |format|
      format.js
      
      format.json do
        render :json => {
          :activities_html => render_to_string({
            :partial => 'activities', :formats => 'html', :locals => {:activities => @unread_activities}
          })
        }, :status => 200
      end
    end
  end
  
  # this action called on mouse hove in dashboard to change the currently
  # shown item in the message view
  def cycle_items
    @message = current_user.total_messages.find params[:id]
    current_index = @message.items.map(&:id).index(params[:current_item].to_i)
    itmes_count = @message.items.count
    next_index = current_index ? ((current_index+1) % itmes_count) : 0
    @item = @message.items[next_index]
    
    respond_to do |format|

      format.json do
        render :json => {
          :dashboard_item_html => render_to_string({
            :partial => 'dashboard_item',
            :formats => 'html',
            :locals => {item: @item, message: @message,
                        new_activities_count: current_user.unread_activities_count(@message)}
          }),
          :new_url => cycle_iems_message_path(:id => @message, :current_item => @item.id)
        }, :status => 200
      end

    end
  end

  def widget
    @widget_version = params[:version]
    
    @collection = Message.public_ones.find_by_identifier!( params[:message_id] )
    
    impression = @collection.impressions.where(user_id: current_user.try(:id)).first_or_initialize
    impression.widget_impressions += 1
    impression.last_viewed_at = Time.now
    impression.last_view_source = 'widget'
    impression.save!
    
    respond_to do |format|
      format.html { render "messages/widget/v#{ @widget_version }/show", :layout => false }
    end
  end

  #-------#
  protected
  #-------#
  
  def authorize_user!
    @message = Message.find( params[:id] )
    
    if ['send_invitation'].include?( params[:action] )
      return true if (@message.private? && current_user.total_messages.include?( @message ))
      return true if (@message.public? && current_user.owned_messages.include?( @message ))
    elsif params[:action] == 'add_item'
      return true if current_user.can_post_item?( @message )
    elsif params[:action] == 'add_comment'
      return true if (@message.public? && @message.add_comment_enabled?)
      return true if (current_user.total_messages.include?( @message ) && @message.add_comment_enabled?)
    elsif ['edit', 'update'].include?( params[:action] )
      return true if (@message.owner_id == current_user.try(:id))
    elsif params[:action] == 'show'
      return true if @message.public?
      
      if current_user
        if @message.personal?
          return true if @message.owner_id == current_user.id
        else
          # Shared Message
          
          @users = @message.total_users
          return true if @users.collect(&:id).include?(current_user.id)
        end
      else
        authenticate_user!
        return false
      end
    else
      return true if @message.public?
      return true if current_user.total_messages.include?( @message )
    end
    
    respond_to do |format|
      format.html { redirect_to root_path }
      format.js { render :json => {}, :status => 403 }
      format.json { render :json => {}, :status => 403 }
    end
  end
  
  def ensure_owner_or_recommender
    @message = Message.find( params[:id] )
    @item = @message.items.find( params[:item] )
    
    return true if (@message.owner_id == current_user.id) || (@item.user_id == current_user.id)
    
    respond_to do |format|
      format.json do
        render :json => {}, :status => 403
      end
    end
  end
  
  def truncated_message_description(description)
    view_context.truncate(description, :length => 150)
  end
end