class Item < ActiveRecord::Base
  default_scope where(deleted: false)

  #-Paperclip-----
    has_attached_file :image, PAPERCLIP_OPTIONS.merge({
      styles: {
        size242: '242x365>', size170: '170x170>', size152: '152x190>', size70: '70x88>',
        medium: "150x150>", thumb: "54x54>",
        size60: '60x60#', size200: '200x200#', size127: '127x127#', size450: '450', size306: '306', size300: '300x300#'
      },
      :default_url => "/assets/item/default.png"
    })
    
    if configatron.paperclip.async_thumbnail_generation
      process_in_background :image, :url_with_processing => false
    end
    
    def image_url(style = nil)
      if self.image_processing?
        self.image.url
      else
        self.image.url(style)
      end
    end
    
    def image_url=(_image_url)
      return if _image_url.blank?
      
      begin
        if _image_url.starts_with?('data:')
          data = _image_url.split(',', 2)[-1]
          StringIO.open(Base64.strict_decode64(data)) do |data|
            data.class.class_eval { attr_accessor :original_filename, :content_type }
            data.original_filename = "temp#{ DateTime.now.to_i }.png"
            data.content_type = "image/png"
            self.image = data
          end
        else
          self.image = self.class.image_url_to_io(_image_url)
        end
      rescue Exception => e
        unless configatron.items.accept_empty_image
          raise(e)
        end
      end
    end
    
    def self.image_url_to_io(image_url)
      require 'open-uri'
      require 'open_uri_redirections'
      
      open(
        URI.encode(image_url), "User-Agent" => "Tisket/1.0 (http://www.tisket.com/)",
        :ssl_verify_mode => OpenSSL::SSL::VERIFY_NONE, :allow_redirections => :safe
      )
    end
  #-----
  
  #-Attributes-----
    attr_accessible :brand, :description, :price, :hide_price, :title, :image, :image_url, :url, :comment,
                    :message_id, :message_attributes, :recipients

    attr_accessor :recipients
    attr_accessor :add_to_new_message
    
    def url_domain
      require_dependency 'url'
      Url.domain(self.url)
    end

    # The URL we're using to get affiliate fees
    def outgoing_url
      self.skimlinks_url
    end

    def skimlinks_url
      url = ERB::Util.url_encode(self.url)
      "http://go.redirectingat.com/?id=57558X1347129&xs=1&url=#{url}&xcust=my_user&sref=http%3A%2F%2Ftisket.com&xs=1"
    end
  #-----
  
  #-Associations-----
    belongs_to :user
    
    belongs_to :message
    accepts_nested_attributes_for :message
    
    has_many :activities
    has_many :comments, class_name: 'Activity', conditions: {kind: Activity::COMMENT}
    
    has_one :missing_image_notification, dependent: :destroy
  #-----

  #-Validations-----
    before_validation do |item|
      unless item.image_content_type.blank?
        unless item.image_content_type.match(/image\/.+|application\/octet-stream|text\/plain/)
          item.image = nil
        end
      end
    end 
    
    #validates :title, presence: true, if: proc {|i| i.processed? }
    validates :user_id, presence: true
    
    validates_presence_of :url
    validate :validate_url
    
    validates_attachment :image, content_type: {
      content_type: /image|application\/octet-stream|text\/plain/ 
    }, allow_blank: true
    
    def valid_url?
      begin
        require 'uri'
        
        # Decode to return to original state then encode to handle non-english urls gracefully
        uri = URI.parse(URI.encode(URI.decode(self.url)))
        
        return false if uri.class == URI::Generic
      rescue URI::InvalidURIError
        return false
      end
      
      return true
    end
    def validate_url
      return if self.url.blank?
      
      unless self.valid_url?
        self.errors.add(:url, 'invalid')
      end
    end
  #-----

  #-Callbacks-----
    before_save :message_presence
    
    # Adjust Message most_recent_item when new item is added or removed (hard remove) or removed (lazy deletion)
    # which is detected by change in deleted flag or resurrected also using deleted flag
    # Same is made with calculation of the message most liked item
    after_create do |item|
      item.message.increment!(:items_count)
      
      item.adjust_message_most_recent_item(item)
      
      # When Item is created with no image, we add missing image notification
      unless item.image.file?
        MissingImageNotification.create!(item_id: item.id)
      end
      
      item.message.update_last_item_created_at!
    end
    after_save do |item|
      if item.deleted_changed?
        item.adjust_message_most_recent_item
        item.adjust_message_most_liked_item
        
        if item.deleted?
          item.missing_image_notification.destroy if item.missing_image_notification
          
          item.message.decrement!(:items_count)
          
          item.remove_from_index!
        else
          item.message.increment!(:items_count)
          
          item.add_to_index! if item.indexable?
        end
        
        item.message.update_items_urls!
      end
      
      # When Item Image is added, we remove the missing image notification
      if item.image_updated_at_changed? && !item.image_updated_at.nil?
        item.missing_image_notification.destroy if item.missing_image_notification
      end
      
      if item.image_updated_at_changed?
        item.message.update_items_urls!
      end
    end
    after_destroy :adjust_message_most_recent_item, :adjust_message_most_liked_item
    after_destroy do |item|
      item.message.decrement!(:items_count)
      
      item.message.update_items_urls!
    end
    
    after_commit :send_notification_messages, :on => :create
  #-----

  #-Scopes-----
    scope :purchased, where(purchased: true)
    scope :processed, where(processed: true)
  #-----

  def calculated_likes_count
    self.activities.where(kind: Activity::LIKE).count
  end

  def set_likes_count!
    self.likes_count = self.calculated_likes_count
    self.save!
  end

  def move_to_collection!(collection_id)
    self.message_id = collection_id
    self.save!
    
    "move"
  end
  
  def copy_to_collection!(collection_id, user)
    copied_item = Item.new
    copied_item.message_id = collection_id
    copied_item.user_id = user.try(:id)
    
    %w(brand title description price hide_price url user_id).each do |field_name|
      copied_item.send("#{ field_name }=", self.send(field_name))
    end
    
    copied_item.processed = false
    
    copied_item.save!
    Background::ItemImageGrabber.perform(copied_item.id, self.image.url)
    
    "copy"
  end
  
  def move_or_copy_to_collection!(collection_id, user)
    if self.message.owner_id == user.id
      if self.message.public? || self.message.personal?
        if self.user_id == user.id
          self.move_to_collection!(collection_id)
        else
          self.copy_to_collection!(collection_id, user)
        end
      else
        self.copy_to_collection!(collection_id, user)
      end
    else
      self.copy_to_collection!(collection_id, user)
    end
  end
  alias_method :copy_or_move_to_collection!, :move_or_copy_to_collection!
  
  #-------#
  protected
  #-------#
  
  def adjust_message_most_recent_item(item = nil)
    self.message.set_most_recent_item!(item)
  end
  def adjust_message_most_liked_item
    self.message.set_most_liked_item!
  end

  def send_notification_messages
    return if self.message.public?
    
    if add_to_new_message
      message.send_invitation_email(user, recipients, self)
    # else
      # recipients = (self.message.total_users - [self.user])
      # recipients = recipients.select {|recipient|
        # (recipient.new_item_notifications? || recipient.all_notifications?)
      # }
#       
      # new_users = []
      # existing_users = []
      # recipients.each do |recipient|
        # if recipient.invited_to_sign_up?
          # new_users << recipient
        # else
          # existing_users << recipient
        # end
      # end
#       
      # if configatron.emails.async
        # Mailer.delay.new_item(self.user, existing_users, self.message, self)
      # else
        # Mailer.new_item(self.user, existing_users, self.message, self).deliver
      # end
      
      # new_users.each do |recipient|
        # if configatron.emails.async
          # Mailer.delay.new_item(self.user, [recipient], self.message, self, true)
        # else
          # Mailer.new_item(self.user, [recipient], self.message, self, true).deliver
        # end
      # end
    end
  end

  def message_presence
    unless message_id.present?
      errors.add(:base, I18n.t('item.message_required'))
      return false
    end
    true
  end
  
  #----#
  public
  #----#
  
  #-Search-----
    def add_to_index!
      search_object = ItemsSearch.find_or_initialize_by_item_id(self.id)
      search_object.collection_id = self.message_id
      search_object.title = self.title
      search_object.brand = self.brand
      search_object.collection_created_at = self.message.created_at
      search_object.save!
    end
    
    def remove_from_index!
      ItemsSearch.where(item_id: self.id).delete_all
    end
    
    def self.index_search(q)
      puts "Search Query: #{ q }"
      q = q.to_s.split(' ').collect {|e| "+#{ e }" }.join(' ')
      puts "Modified Search Query: #{ q }"
      
      if ActiveRecord::Base.connection.adapter_name == "Mysql2"
        relation = ItemsSearch.where("MATCH (title,brand) AGAINST ('#{ q }' IN BOOLEAN MODE)").group('collection_id')
      else
        relation = ItemsSearch.select('*')
      end
      
      relation.order('collection_created_at DESC')
    end
    
    def indexable?
      self.message.public? && self.message.image.file?
    end
  #-----
end