class Message < ActiveRecord::Base
  #-Paperclip-----
    has_attached_file :image, PAPERCLIP_OPTIONS.merge({
      default_url: "/assets/defaults/collection/default.png",
      styles: {
        size300: '300x300#',
        size400: '400x400#'
      }
    })
  #-----
  
  #-Attributes-----
    attr_accessible :subject, :original_content, :active, :public, :image, :url, :remove_image,
                    :add_product_enabled, :add_comment_enabled, :hover_over_enabled, :user_ids, :show_in_explore_page
    
    serialize :info, Hash
    
    def private?
      !self.public?
    end
    
    # def impressions_count
      # self.impressions.sum('site_impressions + widget_impressions').to_i
    # end
    def site_impressions_count
      self.impressions.sum('site_impressions').to_i
    end
    def widget_impressions_count
      self.impressions.sum('widget_impressions').to_i
    end
    
    def cached_impressions_count
      Rails.cache.fetch("message_#{ self.id }_impressions_count", expires_in: 1.hour) do
        self.impressions_count
      end
    end
    def cached_site_impressions_count
      Rails.cache.fetch("message_#{ self.id }_site_impressions_count", expires_in: 1.hour) do
        self.site_impressions_count
      end
    end
    def cached_widget_impressions_count
      Rails.cache.fetch("message_#{ self.id }_widget_impressions_count", expires_in: 1.hour) do
        self.widget_impressions_count
      end
    end
    
    def collection_type_name
      if self.personal?
        'Personal'
      elsif self.shared?
        'Shared'
      else
        'Public'
      end
    end
    
    def url_domain
      require_dependency 'url'
      Url.domain(self.url)
    end
  #-----

  #-Associations-----
    belongs_to :owner, class_name: :User
    
    has_many :message_user_relations, dependent: :destroy
    has_many :users, through: :message_user_relations
  
    has_many :items, dependent: :destroy

    has_many :activities

    has_many :visits, dependent: :destroy
    
    belongs_to :most_recent_item, :class_name => 'Item'
    belongs_to :most_liked_item, :class_name => 'Item'
    
    has_many :impressions, foreign_key: 'collection_id', dependent: :destroy
  #-----

  #-Validations-----
    validates :subject, presence: true
    validates :owner_id, presence: true
    
    validates_uniqueness_of :identifier, :allow_blank => false
    
    validate :validate_url
    
    def valid_url?
      begin
        require 'uri'
        uri = URI.parse(self.url)
        
        return false if uri.class == URI::Generic
      rescue URI::InvalidURIError
        return false
      end
      
      return true
    end
    def validate_url
      return if self.url.blank?
      
      unless self.valid_url?
        self.errors.add(:url, 'invalid')
      end
    end
  #-----

  #-Scopes-----
    scope :active, where(active: true)
    scope :inactive, where(active: false)
    
    scope :public_ones, where(public: true)
    scope :private_ones, where(public: false)
    
    scope :non_featured_ones, where(featured: false)
    scope :featured_ones, where(featured: true)
    
    scope :has_items, where("items_count > 0")
    scope :has_image, where("image_file_name IS NOT NULL")
    scope :recent, order('created_at DESC')
    
    scope :x_days_old, lambda {|days_count|
      where(['created_at > ?', (Time.now.utc.beginning_of_day - days_count.days)])
    }
    
    scope :show_in_explore_page, where(show_in_explore_page: true)
    
    def personal?
      self.private? && (self.users.count == 0)
    end
    
    def shared?
      self.private? && (self.users.count > 0)
    end
  #-----

  #-Callbacks-----
    before_create :update_last_activity_at
    before_create do |record|
      if record.public? && !record.featured?
        record.hover_over_enabled = false
      end
      
      true
    end
    
    before_save :set_closed_at
    before_save do |record|
      record.identifier = SecureRandom.uuid if record.identifier.blank?
    end
    
    after_save do |message|
      if message.public_changed?
        message.adjust_items_index!
      end
    end
  #-----

  def existing_users
    self.users.uniq.select {|user| !user.invited_to_sign_up? }
  end
  def invited_users
    self.users.uniq.select {|user| user.invited_to_sign_up? }
  end

  def set_most_recent_item!(item = nil)
    self.most_recent_item = (item || self.calculated_most_recent_item)
    self.save!
  end
  def calculated_most_recent_item
    self.items.order('created_at desc').first
  end
  
  def set_most_liked_item!
    self.most_liked_item_id = self.calculated_most_liked_item_id
    self.save!
  end

  def total_users
    if owner
      [owner] + users
    else
      users
    end
  end

  def total_users_ids
    self.total_users.collect(&:id)
  end

  def has_user?(user_id_or_record)
    if user_id_or_record.is_a?(User)
      self.total_users.include?( user_id_or_record )
    else
      self.total_users_ids.include?( user_id_or_record )
    end
  end

  def send_invitation_email(inviter, recipients, item = nil)
    return if recipients.empty?
    
    # divide recipients into exsiting users (invite them to the message)
    # and new users (invite them to the system and the message)
    existing_users = []
    new_users = []
    
    new_users_emails = []
    recipients.each do |r|
      if u = User.registered.find_by_email(r)
        # make sure existing users are not in the message before inviting them 
        unless u.total_messages.include?(self)
          existing_users << u
          u.involved_messages << self
          
          # Add inviter as friend to the invitee list of friends
          u.friends << inviter unless u.total_friends.include?(inviter)
        end
      else
        new_users_emails << r
      end
    end
    
    new_users_emails.each do |email|
      u = User.invite!({:email => email}, inviter){|u| u.skip_invitation = true }
      if u.errors.empty?
        u.update_attribute(:invitation_sent_at, Time.now.utc)
        u.involved_messages << self
        
        # Add inviter as friend to the invitee list of friends
        u.friends << inviter unless u.total_friends.include?(inviter)
        
        new_users << u
      end
    end
    
    if item.nil?
      new_users.each do |user|
        if configatron.emails.async
          Mailer.delay.invitation(inviter, [user], self, true)
        else
          Mailer.invitation(inviter, [user], self, true).deliver
        end
      end
      
      if configatron.emails.async
        Mailer.delay.invitation(inviter, existing_users, self)
      else
        Mailer.invitation(inviter, existing_users, self).deliver
      end
    else
      # if configatron.emails.async
        # Mailer.delay.new_item_at_new_message_existing_users(self, item, inviter, existing_users.collect(&:email))
      # else
        # Mailer.new_item_at_new_message_existing_users(self, item, inviter, existing_users.collect(&:email)).deliver
      # end
  
      # if configatron.emails.async
        # Mailer.delay.new_item_at_new_message_new_users(self, item, inviter, u)
      # else
        # Mailer.new_item_at_new_message_new_users(self, item, inviter, u).deliver
      # end
      
      recipients = (existing_users + new_users)
      recipients = recipients.select {|recipient|
        (recipient.new_item_notifications? || recipient.all_notifications?)
      }
      
      new_users = []
      existing_users = []
      recipients.each do |recipient|
        if recipient.invited_to_sign_up?
          new_users << recipient
        else
          existing_users << recipient
        end
      end
      
      if configatron.emails.async
        Mailer.delay.new_item(self.owner, existing_users, self, item)
      else
        Mailer.new_item(self.owner, existing_users, self, item).deliver
      end
      
      new_users.each do |recipient|
        if configatron.emails.async
          Mailer.delay.new_item(self.owner, [recipient], self, item, true)
        else
          Mailer.new_item(self.owner, [recipient], self, item, true).deliver
        end
      end
      
      # if configatron.emails.async
        # Mailer.delay.new_item(inviter, recipients, self, item)
      # else
        # Mailer.new_item(inviter, recipients, self, item).deliver
      # end
    end
    
    # after the above call, each recipients has a record in DB
    # create an activity that the user has been invited to the message
    activities = []
    # (existing_users + new_users).each do |user|
      # #name = user.name.present? ? user.name : user.email
      # # activities << user.activities.create(
        # # message_id: self.id, kind: Activity::SEND_INVITATION,
        # # activity: I18n.t('messages.user_invited', name: name),
        # # performed_by: inviter,
        # # visible: false
      # # )
    # end
    
    activities
  end

  def calculated_most_liked_item_id
    self.activities.likes.joins(:item).where("#{ Item.table_name }.deleted" => false).
         group(:item_id).select('item_id, count(*) as count').order('count desc').first.try(:item_id)
  end
  def calculated_most_liked_item
    Item.find_by_id(self.calculated_most_liked_item_id)
  end

  def purchased_items
    items.where(purchased: true)
  end

  #-Public Collections-----
    def clone_public
      # Transfer collection Details
      pc = self.class.new({
        :subject => self.subject,
        :original_content => self.original_content
      })
      
      # Transfer owner
      pc.owner_id = self.owner_id
      
      # Make it active and public
      pc.active = false
      pc.public = true
      pc.featured = true
      
      pc.add_product_enabled = pc.add_comment_enabled = pc.hover_over_enabled = false
      
      pc.save!
      Background::MessageCloner.perform(self.id, pc.id)
      
      pc
    end
    
    def transfer_items_and_activate!(message_id)
      message = Message.find(message_id)
      
      # Transfer items
      message.items.each do |item|
        new_item = self.items.build
        %w(brand title description price hide_price url user_id).each do |field_name|
          new_item.send("#{ field_name }=", item.send(field_name))
        end
        require 'open-uri'
        new_item.image = open(item.image.url)
        new_item.save!
      end
      
      self.active = true
      self.save!
      
      self
    end
  #-----

  def update_default_image!(force = false)
    return if (self.image.file? && !force)
    
    first_item = self.items.processed.first
    
    if first_item && first_item.image.file?
      self.image = first_item.image
      self.save!
    end
  end

  #-ActiveAdmin - Remove Image Option-----
    attr_accessible :remove_image
  
    attr_writer :remove_image
    def remove_image
      @remove_image || false
    end
    
    before_validation { self.image.clear if self.remove_image == '1' }
  #-----

  def update_items_urls!
    self.info[:items_urls] = []
    self.items.order('created_at ASC').limit(4).each do |i|
      self.info[:items_urls] << i.image.url if i.image.file?  
    end
    self.save!
  end
  
  def items_urls
    self.info[:items_urls] || []
  end

  def remove_image=(remove)
    if remove == '1'
      self.image = nil
    end
  end

  def update_last_item_created_at!
    self.last_item_created_at = self.items.order('created_at DESC').first.try(:created_at)
    self.save!
  end

  #-Search-----
    def adjust_items_index!
      if self.public?
        self.items.each {|item| item.add_to_index! }
      else
        self.items.each {|item| item.remove_from_index! }
      end
    end
  #-----

  #-------#
  protected
  #-------#
  
  def update_last_activity_at
    self.last_activity_at = Time.now
  end
  
  def set_closed_at
    if active_changed? && active == false
      self.closed_at =  Time.now
    end  
  end
end