class User < ActiveRecord::Base
  has_attached_file :avatar, PAPERCLIP_OPTIONS.merge({
    styles: {
      size84: '84x84#', size50: '50x50#', size60: '60x60#', size20: '20x20#', medium: "150x150#",thumb: "54x54#",
      size40: '40x40#', size100: '100x100#'
    },
    :default_url => "/assets/user/:style.jpg"
  })
    
  #-Constants-----
  	AGE_RANGE = ["18 - 23", "24 - 29", "30 - 34", "35 - 40", "41+"]
  	GENDER = %w( Male Female )

    NEVER_FREQUENCY = 0
    NEW_ITEM_FREQUENCY = 1
    ALL_ACTIVITIES_FREQUENCY = 2
	#-----
	
  # Include default devise modules. Others available are:
  # :token_authenticatable, :confirmable,
  # :lockable, :timeoutable and :omniauthable
  devise :invitable, :database_authenticatable, :registerable,
         :recoverable, :rememberable, :trackable, :validatable,
         :omniauthable, :omniauth_providers => [:google, :facebook]

  #-Attributes-----
    # Setup accessible (or protected) attributes for your model
    attr_accessible :email, :age_range, :gender, :first_name, :last_name,
    								:avatar, :password, :password_confirmation, :remember_me,
                    :notification_frequency, :nickname, :location, :website

    def initial(str)
      result = []
      
      unless str.blank?
        result << str[0]
        if str.length > 1
          result << '.'
        end
      end
      
      result.join('') 
    end
    def first_name_initial
      initial(self.first_name)
    end
    def last_name_initial
      initial(self.last_name)
    end
                    
    def initials
      [self.first_name, self.last_name].select {|p| !p.blank? }.collect {|p| p[0] }.join('')
    end
    
    def name
      [self.first_name, self.last_name].select {|p| !p.blank? }.join(' ')
    end
    
    def short_name
      [self.first_name, self.last_name_initial].select {|p| !p.blank? }.join(' ')
    end
    
    def name_or_email
      self.name.presence || self.email
    end
    
    def nickname_or_name
      self.nickname.presence || self.name
    end
    
    def all_friends
      (self.friends + self.reverse_friends).select {|f| !f.invited_to_sign_up? }.uniq.sort_by!(&:first_name)
    end
  #-----
  
  #-Associations-----
    has_one :draft_item, dependent: :destroy
  
    has_many :items, dependent: :nullify
  
    has_many :friendships, foreign_key: :user_id, dependent: :destroy
    has_many :friends, through: :friendships, source: :friend
    has_many :reverse_friendships, class_name: :Friendship, foreign_key: :friend_id, dependent: :destroy
    has_many :reverse_friends, through: :reverse_friendships, source: :user
  
    has_many :message_user_relations, dependent: :destroy
    has_many :involved_messages, through: :message_user_relations
  
    has_many :owned_messages, class_name: :Message, foreign_key: :owner_id, dependent: :destroy

    has_many :search_users, :dependent => :destroy
    
    has_many :activities, :dependent => :destroy
    has_many :activities_likes, :class_name => 'ActivityLike', :dependent => :destroy

    has_many :visits, dependent: :destroy
    
    has_many :authentications, :class_name => 'UserAuthentication', :dependent => :destroy
    
    has_many :impressions, dependent: :nullify
  #-----

  #-Validations-----
    validates :first_name, :last_name, :presence => true
    
    validates_inclusion_of :notification_frequency, :in => [
      NEVER_FREQUENCY, NEW_ITEM_FREQUENCY, ALL_ACTIVITIES_FREQUENCY
    ]
    
    validate :validate_url
    
    def valid_url?
      begin
        require 'uri'
        uri = URI.parse(self.website)
        
        return false if uri.class == URI::Generic
      rescue URI::InvalidURIError
        return false
      end
      
      return true
    end
    def validate_url
      return if self.website.blank?
      
      unless self.valid_url?
        self.errors.add(:website, 'Issue with the website address. Please double check, and try again.')
      end
    end
  #-----

  #-Callbacks-----
    after_save :update_search_user
  #-----

  #-Scopes-----
    scope :all_activities_listner, where(notification_frequency: ALL_ACTIVITIES_FREQUENCY)
    
    scope :requested_access, where(:invitation_sent_at => nil).where("invitation_token IS NOT NULL")
    scope :invited, where("invitation_sent_at IS NOT NULL")
    scope :invited_without_registration, invited.where(:invitation_accepted_at => nil)
    scope :invited_and_registered, invited.where("invitation_accepted_at IS NOT NULL")
    scope :registered, where(:invitation_token => nil)
    scope :directly_registered, registered.where(:invitation_accepted_at => nil)
    
    scope :active_in_past_day, ->{ where('last_sign_in_at >= ?', 1.day.ago.beginning_of_day).where('last_sign_in_at <= ?', 1.day.ago.end_of_day) }
    scope :active_in_past_week, ->{ where('last_sign_in_at >= ?', 1.week.ago.beginning_of_week).where('last_sign_in_at <= ?', 1.week.ago.end_of_week) }
    scope :active_in_past_month, ->{ where('last_sign_in_at >= ?', 1.month.ago.beginning_of_month).where('last_sign_in_at <= ?', 1.month.ago.end_of_month) }
    
    def self.tisket_user
      User.find_by_email('atisketatasket@mailinator.com')
    end
    def self.cached_tisket_user
      Rails.cache.fetch('user_tisket_user') do
        self.tisket_user
      end
    end
  #-----

  def total_friends
    User.where("id in (select friend_id from friendships where user_id = #{id}) or id in (select user_id from friendships where friend_id = #{id})")
  end

  def total_messages(skip_friends = false)
    conditions = [
      "(#{ Message.table_name }.owner_id = #{ self.id })",
      "(#{ Message.table_name }.id in (#{ MessageUserRelation.select('message_id').where(:user_id => self.id).to_sql }))"
    ]
    unless skip_friends
      friends_ids = self.total_friends.pluck('id')
      #friends_ids << self.class.cached_tisket_user.id
      if friends_ids.length > 0
        conditions << "(#{ Message.table_name }.public = #{ true } AND #{ Message.table_name }.owner_id IN (#{ friends_ids.join(',') }))"
      end
    end
        
    Message.where(conditions.join(' OR '))
  end

  def navigation_collections
    self.total_messages(true).active.order('last_item_created_at DESC')
  end

  #-Contacts Search-----
    def terms_for(prefix)
      # Use Hash to avoid duplicates
      all_tags = Hash.new {|h,k| h[k] = {:name => nil, :friend => false, :tisket => false} }
      
      # Loop over local entries which are friends user invited before
      # Add names and mark that they are friends and also tisket users
      local_terms_for(prefix).each do |e|
        all_tags[ e[:email] ][:name] = e[:name].presence
        all_tags[ e[:email] ][:friend] = true
        all_tags[ e[:email] ][:tisket] = true
      end
      
      # Loop over google entries which are not friends and might be tisket users
      google_terms_for(prefix).each do |e|
        all_tags[ e[:email] ][:name] ||= e[:name].presence
        all_tags[ e[:email] ][:friend] ||= false
        all_tags[ e[:email] ][:tisket] ||= false
      end
      
      # Grab all emails that lack names or are not tisket users
      users_emails = []
      all_tags.each_pair do |email, info|
        if info[:name].blank? || info[:tisket] == false
          users_emails << email
        end
      end
      
      # Grab users with these emails to take names from them and to set them as tisket users
      users = User.where(:email => users_emails)
      users.each do |user|
        all_tags[ user.email ][:name] = user.name unless user.name.blank?
        all_tags[ user.email ][:tisket] = true
      end
      
      # Loop and order entries based on their presence in one of three
      friends = [] ; tisket_users = [] ; others = []
      all_tags.each_pair do |email, info|
        if info[:friend]
          if info[:name].blank?
            friends << {
              label: "#{ email } (Tisket User)",
              value: email
            }
          else
            friends << {
              label: "#{ info[:name] } (Tisket User)",
              value: email
            }
          end
        elsif info[:tisket]
          if info[:name].blank?
            tisket_users << {
              label: "#{ email } (Tisket User)",
              value: email
            }
          else
            tisket_users << {
              label: "#{ info[:name] } (Tisket User)",
              value: email
            }
          end
        else
          if info[:name].blank?
            others << {
              label: email,
              value: email
            }
          else
            others << {
              label: "#{ info[:name] } (#{ email })",
              value: email
            }
          end
        end
      end
      
      (friends + tisket_users + others)
    end
    
    def local_terms_for(prefix)
      suggestions = search_users.where("term like ?", "%#{prefix.downcase}%").select([:name, :email])  
      suggestions.map{|s| {name: s.name.rstrip.lstrip, email: s.email} }
    end
    
    def google_terms_for(prefix)
      return [] if self.authentications.google.empty?
      
      # Search on Google
      begin
        search_google_contacts(prefix)
      rescue GContacts::Unauthorized => e
        begin
          google_authentication = self.authentications.google.first
          
          if e.message == "Token revoked"
            Rails.logger.debug ">>>>>>>>>>>>>> Token Revoked"
            google_authentication.destroy if google_authentication
          else  
            response = OAuth2::AccessToken.new(
              OAuth2::Client.new(configatron.google.client_id, configatron.google.client_secret, {
                :site => 'https://accounts.google.com', :authorize_url => "/o/oauth2/auth",
                :token_url => "/o/oauth2/token"
              }),
              google_authentication.info[:token],
              :refresh_token => google_authentication.info[:refresh_token]
            ).refresh!
            
            google_authentication.info[:token] = response.token
            google_authentication.save!
            
            search_google_contacts(prefix)
          end
        rescue Exception => e2
          Airbrake.notify_or_ignore(e2, {
            :parameters => {
              :user_id => self.id
            }
          })
          
          google_authentication.destroy if google_authentication
          
          []
        end
      rescue Exception => e
        Airbrake.notify_or_ignore(e, {
          :parameters => {
            :user_id => self.id
          }
        })
        
        []
      end
    end
    
    def search_google_contacts(query)
      require_dependency 'google_contacts'
        
      google_authentication = self.authentications.first
      client = GContacts::Client.new(:access_token => google_authentication.info[:token])
      
      entries = []
      
      client.all(:params => {'q' => query, 'max-results' => 10}).each_with_index do |entry, index|
        puts ">>>>>>>> #{ index}"
        
        (entry.data["gd:email"] || []).each {|e|
          entries << {
            :name => entry.title,
            :email => e["@address"]
          }
        }
      end
      
      entries
    end
  #-----

  def like(message_id_or_record, item_id)
    if message_id_or_record.is_a?(Message)
      @message = message_id_or_record
    else
      @message = self.total_messages.find( message_id_or_record )
    end
    
    @item = @message.items.find( item_id )
    @like = activities.find_or_initialize_by_kind_and_message_id_and_item_id(Activity::LIKE, @message.id, item_id)
    
    if @like.new_record?
      @like.activity = I18n.t('messages.like.activity_message', product: @item.title,
                                                                brand: @item.brand,
                                                                name: self.first_name)
      @like.performed_by = self
      @like.visible = false
      @like.save
      
      @item.increment!(:likes_count)
      @message.set_most_liked_item!
    end

    self.unleave(@message, item_id)

    [@message, @item, @like]
  end

  def unlike(message_id_or_record, item_id)
    if message_id_or_record.is_a?(Message)
      @message = message_id_or_record
    else
      @message = self.total_messages.find( message_id_or_record )
    end
    
    @item = @message.items.find( item_id )
    @like = activities.find_by_kind_and_message_id_and_item_id(Activity::LIKE, @message.id, @item.id)
    return if @like.nil?
    @activity_id = @like.id
    
    @like.destroy
    @item.decrement!(:likes_count)
    @message.set_most_liked_item!
    
    [@message, @item, @activity_id]
  end

  def leave(message_id_or_record, item_id)
    if message_id_or_record.is_a?(Message)
      @message = message_id_or_record
    else
      @message = self.total_messages.find( message_id_or_record )
    end
    
    @item = @message.items.find item_id
    @like = activities.find_or_initialize_by_kind_and_message_id_and_item_id(Activity::LEAVE, @message.id, item_id)
    
    if @like.new_record?
      @like.activity = I18n.t('messages.like.activity_message', product: @item.title,
                                                                brand: @item.brand,
                                                                name: self.first_name)
      @like.performed_by = self
      @like.visible = false
      @like.save
      
      @item.increment!(:leaves_count)
      #@message.set_most_liked_item!
    end

    self.unlike(@message, item_id)

    [@message, @item, @like]
  end

  def unleave(message_id_or_record, item_id)
    if message_id_or_record.is_a?(Message)
      @message = message_id_or_record
    else
      @message = self.total_messages.find( message_id_or_record )
    end
    
    @item = @message.items.find item_id
    @like = activities.find_by_kind_and_message_id_and_item_id(Activity::LEAVE, @message.id, @item.id)
    return if @like.nil?
    @activity_id = @like.id
    
    @like.destroy
    @item.decrement!(:leaves_count)
    #@message.set_most_liked_item!
    
    [@message, @item, @activity_id]
  end

  def like?(message_id, item_id)
    activities.find_by_kind_and_message_id_and_item_id(Activity::LIKE, message_id, item_id)
  end 
  def hate?(message_id, item_id)
    activities.find_by_kind_and_message_id_and_item_id(Activity::LEAVE, message_id, item_id)
  end

  def like_activity?(activity)
    self.activities_likes.exists?(:activity_id => activity.id)
  end

  def last_visit(message_id)
    self.visits.where(message_id: message_id).first.try(:last_visit)
  end

  def unread_activities(message)
    last_visit_date = last_visit(message.id)
    
    if last_visit_date
      message.activities.where("created_at > ? and user_id != ?",
                                last_visit_date, self.id)
    else
      message.activities.where("user_id != ?", self.id)
    end
  end

  def unread_activities_ids(message)
    last_visit_date = last_visit(message.id)
    
    if last_visit_date
      message.activities.where("created_at >= ? and user_id != ?",
                              last_visit_date, self.id).select(:id).map(&:id)
    else
      message.activities.where("user_id != ?", self.id).select(:id).map(&:id)
    end
  end

  def unread_activities_count(message)
    unread_activities = message.activities.where("user_id != ?", self.id)
    
    last_visit_date = last_visit(message.id)
    if last_visit_date
      unread_activities = unread_activities.where("created_at >= ?", last_visit_date)
    end
    
    unread_activities.count
  end

  def no_notifications?
    notification_frequency == NEVER_FREQUENCY
  end

  def new_item_notifications?
    notification_frequency == NEW_ITEM_FREQUENCY
  end

  def all_notifications?
    notification_frequency == ALL_ACTIVITIES_FREQUENCY
  end
  
  def visit_message(message, visit = nil)
    visit ||= self.visits.find_or_initialize_by_message_id(message.id)
    
    visit.last_visit = Time.now
    visit.activities_count = 0
    visit.save
  end
  def visit_message!(message)
    raise unless visit_message(message)
  end
  
  def daily_summary
    digest = []
    purchased_items = []

    self.total_messages.each do |msg|
      activities = self.unread_activities(msg)
      if activities.present?
        new_items = activities.new_items
        likes = activities.likes.group(:item_id).count
        comments = activities.comments
        
        first_comment = "\"#{comments[0].activity}\" by #{comments[0].user.name}" if comments[0]
        second_comment = "\"#{comments[1].activity}\" by #{comments[1].user.name}" if comments[1]
        
        if likes.keys[0]
          first_liked_item = activities.likes.where(item_id: likes.keys[0]).first.item
          first_liked_item_likes_count = likes[likes.keys[0]] 
        end

        if likes.keys[1]
          second_liked_item = activities.likes.where(item_id: likes.keys[1]).first.item
          second_liked_item_likes_count = likes[likes.keys[1]] 
        end
        
        item = if new_items.present?
          new_items.first.item
        else
          msg.most_liked_item
        end

        digest << {
          message: msg,
          new_items_count: new_items.count,
          likes_count: activities.likes.count,
          comments_count: comments.count,
          first_comment: first_comment,
          second_comment: second_comment,
          first_liked_item: first_liked_item,
          first_liked_item_likes_count: first_liked_item_likes_count,
          second_liked_item: second_liked_item,
          second_liked_item_likes_count: second_liked_item_likes_count,
          item: item
         }
      end
      
      unless msg.active?
        last_visit = self.last_visit(msg.id)
        if last_visit.nil? || msg.closed_at >= last_visit
          msg.items.where(purchased: true).each do |item|
            purchased_items << item
          end
        end
      end
    end
    
    {digest: digest, purchased_items: purchased_items}
  end
  
  #-Omniauth-----
    def password_required?
      return true if self.encrypted_password_changed?
      
      self.authentications.empty? && self.encrypted_password.blank? && super
    end
  
    def password_changed?
      self.encrypted_password_changed?
    end
  
    def apply_omniauth(auth)
      if auth.provider == 'google'
        self.email = auth.extra.raw_info.email if self.email.blank?
        self.first_name = auth.extra.raw_info.given_name if self.first_name.blank?
        self.last_name = auth.extra.raw_info.family_name if self.last_name.blank?
        self.gender = auth.extra.raw_info.gender.capitalize if self.gender.blank?
        
        self.fetch_picture(auth, 'google') unless self.avatar.file?

        self.authentications.build({
          :provider => auth.provider, :uid => auth.uid, :info => {
            :token => auth.credentials.token,
            :refresh_token => auth.credentials.refresh_token
          }
        })
      elsif auth.provider == 'facebook'
        # User provider email if email not set or if this user is invited user
        self.email = auth.info.email if self.email.blank? || self.invitation_token.present?
        self.first_name = auth.info.first_name if self.first_name.blank?
        self.last_name = auth.info.last_name if self.last_name.blank?
        self.gender = auth.info.gender.capitalize if auth.info.gender and self.gender.blank?

        self.fetch_picture(auth, 'facebook') unless self.avatar.file?

        self.authentications.build({
          :provider => auth.provider, :uid => auth.uid, :info => {
            :token => auth.credentials.token
          }
        })
      end
    end

    def fetch_picture(auth, provider)
      if provider == 'google'
        image_url = auth.extra.raw_info.picture
      elsif provider == 'facebook'
        # Changing protocol from http to https to avoid raise of exception from open-uri since the url called
        # will be redirected to another with different protocol which is "https"
        # Note: Don't remove unless you know what you are doing
        image_url = auth.info.image.gsub('http://', 'https://')
        image_url += "?type=large" unless image_url.blank?
      end
        
      return if image_url.blank?
      
      self.avatar = open(image_url)
    end
  #-----
  
  def connected_to_google?
    self.authentications.google.count > 0
  end
  
  def can_edit?(item)
    (item.user_id == self.id || item.message.owner_id == self.id)
  end
  
  def can_post_item?(message)
    return false unless message.add_product_enabled?
    
    if (message.public? || message.personal?)
      return (message.owner_id == self.id)
    else
      return self.total_messages(true).active.include?( message )
    end
    
    return false
  end
  
  #-Statistics-----
    def self.active_weeks
      response = []
      
      time_now = Time.now.utc
      
      first_user = User.order('created_at ASC').first
      start_time = first_user.created_at.beginning_of_week
      end_time = first_user.created_at.end_of_week
      while(true)
        break if time_now < start_time
        
        response << [start_time, end_time]
        
        start_time = end_time + 1
        end_time = start_time.end_of_week
      end
      
      response
    end
    
    def self.active_months
      response = []
      
      time_now = Time.now.utc
      
      first_user = User.order('created_at ASC').first
      start_time = first_user.created_at.beginning_of_month
      end_time = first_user.created_at.end_of_month
      while(true)
        break if time_now < start_time
        
        response << [start_time, end_time]
        
        start_time = end_time + 1
        end_time = start_time.end_of_month
      end
      
      response
    end
    
    def self.results_per_active_months
      response = []  
    
      statistics = {}
      UsersStatistic.all.each do |record|
        statistics["#{ record.start_time }-#{ record.end_time }"] = record.info
      end
    
      time_now = Time.now.utc
    
      self.active_months.each do |active_month|
        info = {} 
        
        if (record = statistics["#{ active_month[0] }-#{ active_month[1] }"])
          info = record
        else
          begin
            require_dependency 'google_analytics/importer'
            google_analytics_data = GoogleAnalytics::Importer.new.run({
              :start_date => active_month[0].to_date.to_s,
              :end_date => active_month[1].to_date.to_s
            })
            
            info[:active_users] = (google_analytics_data["do_activity"] || 0)
            info[:active_users_with_collections] = (google_analytics_data["create_collection"] || 0)
            info[:asked_for_invitation] = self.requested_access.where(
              ["created_at >= ? AND created_at <= ?", active_month[0], active_month[1]]
            ).count
            info[:invited_without_registration] = self.invited_without_registration.where(
              ["invitation_sent_at >= ? AND invitation_sent_at <= ?", active_month[0], active_month[1]]
            ).count 
            
            invited_and_registered = self.invited_and_registered.where(
              ["invitation_accepted_at >= ? AND invitation_accepted_at <= ?", active_month[0], active_month[1]]
            ) 
            info[:invited_and_registered] = invited_and_registered.length
            
            info[:invited] = (info[:invited_without_registration] + info[:invited_and_registered])
            
            directly_registered = self.directly_registered.where(
              ["created_at >= ? AND created_at <= ?", active_month[0], active_month[1]]
            )
            info[:directly_registered] = directly_registered.length
            
            info[:registered] = (info[:invited_and_registered] + info[:directly_registered])
            info[:registered_through_carousel] = (invited_and_registered + directly_registered).select {|e|
              e.invitation_medium == 'carousel'
            }.length
          
            info[:total_items_added] = Item.where(
              ["created_at >= ? AND created_at <= ?", active_month[0], active_month[1]]
            ).count
            info[:total_collections_created] = Message.where(
              ["created_at >= ? AND created_at <= ?", active_month[0], active_month[1]]
            ).count
          
            unless (time_now >= active_month[0] && time_now <= active_month[1]) 
              UsersStatistic.create!({
                :start_time => active_month[0],
                :end_time => active_month[1],
                :info => info
              })
            end
          rescue Exception => e
            Airbrake.notify_or_ignore(e)
            
            info = {
             :asked_for_invitation => 0,
             :invited => 0,
             :invited_and_registered => 0,
             :directly_registered => 0,
             :registered => 0,
             :registered_through_carousel => 0,
             :active_users => 0,
             :active_users_with_collections => 0,
             :total_items_added => 0,
             :total_collections_created => 0
            }
          end
        end
        
        response << {
          :start_date => active_month[0].to_date,
          :end_date => active_month[1].to_date,
          :info => info
        }
      end
      
      response
    end
  #-----
  
  def facebook_friends
    Bundler.require(:facebook)
    
    begin
      fu = FbGraph::User.new('me', :access_token => self.authentications.facebook.first.info[:token])
      
      fu.friends
    rescue Exception => e
      Airbrake.notify_or_ignore(e)
      
      []
    end
  end
  
  def facebook_friends_users
    User.where(
      id: UserAuthentication.where(uid: self.facebook_friends.collect {|fu|
        fu.raw_attributes["id"]
      }).pluck('user_id')
    )
  end
  
  def store_facebook_friends!
    begin
      self.facebook_friends_users.each do |friend|
        self.friendships.find_or_create_by_user_id_and_friend_id(self.id, friend.id)
      end
    rescue Exception => e
      Airbrake.notify_or_ignore(e)
    end
  end
  
  def saved_items_collection
    self.owned_messages.find_or_create_by_subject('Saved Items')
  end
  
  #-------#
  protected
  #-------#
  
  def update_search_user
    if first_name_changed? || last_name_changed? || email_changed?
      old_email = email_changed? ? email_was : email
      SearchUser.update_all({name: self.name, email: self.email,
                            term: "#{self.name.downcase} #{self.email.downcase}"},
                            {email: old_email})
    end    
  end
end